/* $Id: afsroot.C,v 1.20 1999/12/03 05:03:54 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


#include "sfscd.h"
#include "rxx.h"

afsnode *
afsroot::lookup (const str &name, const svccb *sbp)
{
  afsnode *e = afsdir::lookup (name, sbp);
  if (!e)
    if (afsdir *d = userdir (sbp))
      e = d->lookup (name, sbp);
  return e;
}

void
afsroot::nfs_lookup (svccb *sbp, str name)
{
  afsnode *e = lookup (name, sbp);
  if (!isnfsmounter (sbp->getaup ())) {
    if (ptr<afsnode> r = revocation::lookup (name)) {
      if (e != r) {
	e = r;
	ref<afsdir> d (userdir (sbp));
	d->unlink (name);
	d->link (e, name);
	if (d == afs_naroot)
	  timecb (timenow + 30, wrap (afsdir_unlink, d, name));
      }
    }
    else {
      int err = srvinfo::pathstate (name);
      if (err > 0) {
	nfs_error (sbp, nfsstat (err));
	return;
      }
      if (!e && !terminating)
	e = automount (sbp, name);
    }
  }
  lookup_reply (sbp, e);
}

bool
afsroot::entryok (afsdirentry *de, const svccb *sbp)
{
  afsdir *d = userdir (sbp);
  if (d && d->entryok (de, sbp))
    return true;
  return afsdir::entryok (de, sbp);
}

afsdirentry *
afsroot::firstentry (const svccb *sbp)
{
  afsdir *d = userdir (sbp);
  afsdirentry *de = afsdir::firstentry (sbp);
  if (!d || !d->lookup (de->name, sbp))
    return de;
  return nextentry (de, sbp);
}

afsdirentry *
afsroot::nextentry (afsdirentry *de, const svccb *sbp)
{
  afsdir *d = userdir (sbp);
  if (de->dir == this) {
    for (de = afsdir::nextentry (de, sbp); de;
	 de = afsdir::nextentry (de, sbp))
      if (!d || !d->lookup (de->name, sbp))
	return de;
  }
  else {
    assert (de);
  }

  if (!d)
    return NULL;
  if (!de)
    return d->firstentry (sbp);
  else
    return d->nextentry (de, sbp);
}

void
afsroot::mkfattr3 (fattr3 *f, const svccb *sbp)
{
  /* BSD needs the seconds (not just milliseconds/nanoseconds) of the
   * mtime to change on every lookup/getattr in order to defeat the
   * name cache. */
  const sfs_aid aid = cdgetaid (sbp->getaup ());
  if (aid != lastaid) {
    lastaid = aid;
    bumpmtime ();
  }
  afsdir::mkfattr3 (f, sbp);
  if (afsdir *d = userdir (sbp))
    f->nlink += d->getnlinks () - 2;
}

bool
nameok (const str &name)
{
  static rxx namerx ("[a-zA-Z0-9\\-]+(\\.[a-zA-Z0-9\\-]+)*");
  return name.len () < NFS_MAXNAMLEN && namerx.match (name);
}

bool
afsusrdir::mkulink (const str &path, const str &name)
{
  if (!nameok (name))
    return false;
  unlink (name);
  if (path && path.len ()) {
    negcache.remove (name);
    if (nentries < maxulinks)
      symlink (path, name);
    else {
      // XXX - what to do?
      warn ("afsusrdir: maxulinks exceeded\n");
      return false;
    }
  }
  else {
    if (negcache.size () >= maxulinks)
      // XXX - this is kind of low-tech
      negcache.clear ();
    negcache.insert (name);
  }
  return true;
}

void
afsusrdir::clrulink (const str &name)
{
  negcache.remove (name);
  unlink (name);
  bumpmtime ();
}
